# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/wdika/NeMo/blob/main/nemo/core/utils/process_launcher/launcher.py

import copy
import hashlib
import os
import subprocess
import sys
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence

import torch
from hydra.core.config_store import ConfigStore
from hydra.core.hydra_config import HydraConfig
from hydra.core.plugins import Plugins
from hydra.core.singleton import Singleton
from hydra.core.utils import JobReturn, JobStatus, configure_log, filter_overrides, setup_globals
from hydra.plugins.launcher import Launcher
from hydra.types import HydraContext, TaskFunction
from omegaconf import DictConfig, OmegaConf, open_dict

from atommic.utils import logging


# monkey-patch hydra func
def is_in_toplevel_plugins_module(*args, **kwargs) -> bool:  # pylint: disable=unused-argument
    """Monkey-patch Hydra to allow for plugins to be loaded from any module."""
    return True


# Monkey-patch Hydra
Plugins.instance().is_in_toplevel_plugins_module = is_in_toplevel_plugins_module


@dataclass
class ProcessLauncherConfig:
    """Configuration for the ProcessLauncher plugin."""

    _target_: str = "atommic.core.utils.process_launcher.launcher.ProcessLauncher"
    num_gpus: int = -1
    jobs_per_gpu: int = 1


def execute_job(
    idx: int,
    overrides: Sequence[str],
    hydra_context: HydraContext,
    config: DictConfig,
    singleton_state: Dict[Any, Any],
    gpu_idx: int,
):
    """Creates a process that launches a "single" job that is identical in config + updated with sweep hyperparams.
    Since a different process is being used, CUDA can work in non-ddp mode without issue.
    Attempting ddp when using this script will not work as ddp cannot be used in shared contexts.

    Parameters
    ----------
        idx : int
            Global index of the job.
        overrides : list
            List of str overrides that correspond to this job.
        hydra_context : HydraContext
            Hydra Context used to load the sweep params into the global config.
        config : DictConfig
            Global config that will be updated with sweep hyper parameters.
        singleton_state : Dict[Any, Any]
            Hydra state.
        gpu_idx : int
            The GPU ID on which this process will be run.

    Returns
    -------
    tuple
        - The Process object that corresponds to this sweep.
        - The JobReturn object holding some metadata about this run.
    """
    # Required by Hydra (lookup other Hydra Launchers for details)
    setup_globals()
    Singleton.set_state(singleton_state)

    # Update base config with overrides to create sweep config
    sweep_config = hydra_context.config_loader.load_sweep_config(config, list(overrides))
    with open_dict(sweep_config):
        sweep_config.hydra.job.id = f"{sweep_config.hydra.job.name}_{idx}"
        sweep_config.hydra.job.num = idx
    HydraConfig.instance().set_config(sweep_config)

    # Set up a directory where the config will temporarily be stored.
    script_path = os.path.join(os.getcwd(), sys.argv[0])
    script_path = os.path.abspath(script_path)

    hash_salt = "|".join([script_path, str(OmegaConf.to_yaml(config))]).encode("utf-8")
    hash_val = hashlib.sha256(hash_salt).hexdigest()

    config_dir = os.path.join(os.getcwd(), "hydra_cfg", str(hash_val))
    if not os.path.exists(config_dir):
        os.makedirs(config_dir, exist_ok=True)

    task_cfg = copy.deepcopy(sweep_config)

    # Remove hydra from sweep config
    # This is done to prevent recursive call to multirun launcher in Hydra.
    with open_dict(task_cfg):
        task_cfg.pop("hydra", "")

    # Save the current jobs config to directory
    temp_config_name = f"config_{idx}.yaml"
    temp_config = os.path.join(config_dir, temp_config_name)
    OmegaConf.save(task_cfg, temp_config)

    # Compute the overides as a dict
    overrides = OmegaConf.to_container(config.hydra.overrides.task)

    # Check and replace trainer.devices in config with gpu_idx
    found_devices = False
    gpu_override = f"trainer.devices=[{gpu_idx}]"
    for oidx, val in enumerate(overrides):
        if "trainer.devices" in val:
            overrides[oidx] = gpu_override  # type: ignore
            found_devices = True

    if not found_devices:
        overrides.append(gpu_override)  # type: ignore

    # Build launch command
    # Note: We depend on PTL doing the right thing since this command has global visibility of all CUDA_VISIBLE_DEVICES
    cmd = [
        "python",
        script_path,
        "--config-path",
        config_dir,
        "--config-name",
        temp_config_name,
        *overrides,
    ]

    # Launch the subprocess; pipe the stderr
    # NOTE: If this hangs due to some reason after prolonged training, it means that the stderr pipe buffer
    # has become full at the OS level and we need to explicitly empty it (either parallel thread or manually
    # call proc.communicate(). It should not happen in general case as stderr is filled only in case retcode != 0
    # If it does happen though, implement the code here
    # https://stackoverflow.com/questions/39607172/python-subprocess-popen-poll-seems-to-hang-but-communicate-works
    proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)  # pylint: disable=consider-using-with

    # Setup data thread for stderr
    std_error_buffer: List = []
    # Trivial thread just reads lines from stdout into the list
    drainerthread = threading.Thread(target=std_error_buffer.extend, args=(proc.stderr,))
    drainerthread.daemon = True
    drainerthread.start()

    # Construct JobReturn object for Hydra
    res = JobReturn()
    res.cfg = task_cfg
    res.overrides = overrides
    res.hydra_cfg = config
    res.working_dir = os.getcwd()
    res.return_value = None

    return proc, res, (std_error_buffer, drainerthread)


def launch(  # noqa: MC0001
    launcher, job_overrides: Sequence[Sequence[str]], initial_job_idx: int
) -> Sequence[JobReturn]:
    """Launches a list of jobs with the given config.

    Parameters
    ----------
        launcher : class
            Reference to the Launched subclass.
        job_overrides : list
            A List of List<String>, where each inner list is the arguments for one job run.
        initial_job_idx : int
            Initial job idx in batch.

    Returns
    -------
    list
        A list of JobReturn objects.
    """
    # Needed for Hydra, lookup JoblibLauncher in Hydra
    setup_globals()
    assert launcher.config is not None
    assert launcher.task_function is not None
    assert launcher.hydra_context is not None

    configure_log(launcher.config.hydra.hydra_logging, launcher.config.hydra.verbose)
    sweep_dir = Path(str(launcher.config.hydra.sweep.dir))
    sweep_dir.mkdir(parents=True, exist_ok=True)

    # Extract the runner's config (it is actually a DictConfig, but type is used for autocomplete)
    runner_cfg = launcher.runner  # type: ProcessLauncherConfig

    logging.info(
        f"ProcessLauncher({','.join([f'{k}={v}' for k, v in runner_cfg.items()])}) "  # type: ignore
        f"is launching {len(job_overrides)} jobs."
    )
    logging.info(f"Launching jobs, sweep output dir : {sweep_dir}")
    for idx, overrides in enumerate(job_overrides):
        logging.info(f"\t#{idx} : {' '.join(filter_overrides(overrides))}")

    # Needed by Hydra
    singleton_state = Singleton.get_state()

    # Process the runner's config to build up the multiplex config
    num_gpus = runner_cfg.get("num_gpus", -1)  # type: ignore
    jobs_per_gpu = runner_cfg.get("jobs_per_gpu", 1)  # type: ignore

    # Only GPUs are supported for now.
    if num_gpus <= 0:
        if torch.cuda.is_available():
            num_gpus = torch.cuda.device_count()
        else:
            raise ValueError(f"{launcher.__class__.__name__} only supports GPU operations.")

    # Setup arguments for multiplex runner
    overrides = list(job_overrides)  # type: ignore
    num_overrides = len(overrides)

    job_idx = 0
    batch_size = num_gpus * jobs_per_gpu
    gpu_idx = 0

    ret: List = []  # List of returned JobResult
    subprocess_list: List = []  # Buffer of subprocess
    results = []  # Buffer of JobResult

    # STD ERROR cache
    std_error_buffers = []  # type: List[List[str]]
    std_error_threads: threading.Thread = []  # type: ignore

    # Run over all job combinations
    while job_idx < num_overrides:
        # Fill up subprocess buffer while its size is smaller than multiplex batch size
        while len(subprocess_list) < batch_size:
            # If we run out of jobs, stop trying to submit more jobs
            if job_idx >= num_overrides:
                break

            # Submit a job as a new process
            process, res, error_tup = execute_job(
                initial_job_idx + job_idx,
                overrides[job_idx],
                launcher.hydra_context,
                launcher.config,
                singleton_state,
                gpu_idx % num_gpus,  # This will evenly distribute GPU load
            )

            # Store the subprocesses and JobResults
            subprocess_list.append(process)
            results.append(res)

            # Manage stderror thread data
            std_error_buffers.append(error_tup[0])
            std_error_threads.append(error_tup[1])  # type: ignore

            job_idx += 1
            gpu_idx += 1

        # Poll for samples in batch to finish.
        if len(subprocess_list) > 0:
            finished_processes = [0] * len(subprocess_list)

            # Check if all processes are completed or not
            # This is busy waiting, this is actually quite necessary
            # Turns out that when you do proc.communicate(), you block all other threads immediately.
            # IE they may fill up their buffers entirely, and hang while they wait for the first thread
            # who called communicate() to finish its work or crash.
            # Effectively it entirely stops multiprocessing jobs or multiplexed runs.
            # Must poll and busy wait to keep threads alive, along with drain the pipes with thread buffers.
            while sum(finished_processes) < len(subprocess_list):
                # Check all processes to make sure they have a retcode (doesn't matter yet if 0 or not)
                for proc_idx, proc in enumerate(subprocess_list):
                    # poll() is cheaper op than communicate()
                    retcode = proc.poll()

                    if retcode is not None:
                        # Log that the process with some ID has finished
                        if finished_processes[proc_idx] == 0:
                            logging.info(f"Processed job : {len(ret) + proc_idx} :: Ret code = {retcode}")

                        finished_processes[proc_idx] = 1

                        # Join this thread and merge its stderror buffer
                        proc.wait()
                        std_error_threads[proc_idx].join()  # type: ignore
                        error_data = std_error_buffers[proc_idx]
                        error_data = [
                            str(data, encoding="utf-8").encode("utf-8").decode("utf-8").encode("utf-8")  # type: ignore
                            for data in error_data
                        ]

                        std_error_buffers[proc_idx] = error_data

                time.sleep(1.0)

            # Process all the subprocess results
            for proc_idx, (proc, res) in enumerate(zip(subprocess_list, results)):
                # Wait until completion of process
                _ = proc.communicate()

                # 0 is for successful run
                if proc.returncode == 0:
                    res.status = JobStatus.COMPLETED
                else:
                    # > 0 is for error, log the error.
                    # Note: For the sake of efficiency while we log the error and raise an exception,
                    # It will only raise the 1st wrong job in all the jobs.
                    # If multiple jobs fail, it will still try to execute every job first before
                    # raising the error for the first one.
                    # This is done so that even if some jobs fail (say OOM or something),
                    # other jobs can still run.
                    err_buffer = std_error_buffers[proc_idx]
                    if isinstance(err_buffer, (list, tuple)):
                        err_string = ""
                        for err_line in err_buffer:
                            err_string = (
                                err_string
                                + f"{str(err_line, encoding='utf-8').encode('utf-8').decode('utf-8')}"  # type: ignore
                            )
                    else:
                        err_string = err_buffer

                    error_msg = (
                        f"\nHyperparameter Arguments : {proc.args}\n"
                        f"Process Return code : {proc.returncode}\n"
                        f"Error Trace :\n"
                        f"{err_string}"
                    )
                    res.return_value = Exception(error_msg)
                    res.status = JobStatus.FAILED

                logging.info(f"Finished executing job : {len(ret)}. Return Code = {proc.returncode}")
                ret.append(res)

            # Reset for next batch
            subprocess_list.clear()
            results.clear()

    return ret


class ProcessLauncher(Launcher):
    """Process Launcher

    Based on the JoblibLauncher, but uses processes to scatter jobs in a multiplexed manner across some number of GPUs
    on a single machine.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Inits :class:`ProcessLauncher`."""
        self.config: Optional[DictConfig] = None
        self.task_function: Optional[TaskFunction] = None
        self.hydra_context: Optional[HydraContext] = None
        self.runner: ProcessLauncherConfig = kwargs  # type: ignore

    def setup(
        self,
        *,
        hydra_context: HydraContext,
        task_function: TaskFunction,
        config: DictConfig,
    ) -> None:
        """Setups the launcher.

        Parameters
        ----------
        hydra_context : HydraContext
            Hydra context object.
        task_function : TaskFunction
            Task function.
        config : DictConfig
            Launcher config.
        """
        self.config = config
        self.task_function = task_function
        self.hydra_context = hydra_context

    def launch(self, job_overrides: Sequence[Sequence[str]], initial_job_idx: int) -> Sequence[JobReturn]:
        """Launch jobs.

        Parameters
        ----------
        job_overrides : Sequence[Sequence[str]]
            List of List<String>, where each inner list is the arguments for one job run.
        initial_job_idx : int
            Initial job idx.

        Returns
        -------
        Sequence[JobReturn]
            Sequence of JobReturn objects, where each object represents the return value of one job.
        """
        return launch(launcher=self, job_overrides=job_overrides, initial_job_idx=initial_job_idx)


ConfigStore.instance().store(
    group="hydra/launcher",
    name="atommic_launcher",
    node=ProcessLauncherConfig,
    provider="atommic_process_launcher",
)

Plugins.instance(ProcessLauncher)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import math
from abc import ABC
from typing import List, Tuple

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

import atommic.collections.segmentation.nn.base as base_segmentation_models
from atommic.core.classes.common import typecheck

__all__ = ["BaseSegmentationNet"]


class BaseSegmentationNet(base_segmentation_models.BaseMRISegmentationModel, ABC):
    """Abstract class for all segmentation models."""

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """inits :class:`BaseSegmentationNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.
        trainer : Trainer, optional
            PyTorch Lightning trainer object, by default None.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.padding_size = cfg_dict.get("segmentation_module_padding_size", 11)
        self.normalize = cfg_dict.get("segmentation_module_normalize", False)
        self.norm_groups = cfg_dict.get("segmentation_module_norm_groups", 2)
        self.segmentation_module = self.build_segmentation_module(cfg)

    def build_segmentation_module(self, cfg: DictConfig) -> torch.nn.Module:
        """Build the segmentation module.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object specifying the model's hyperparameters.
        """
        raise NotImplementedError

    def norm(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Normalize the input."""
        # group norm
        b, c, s, h, w = x.shape

        x = x.reshape(b, self.norm_groups, -1)

        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)

        x = (x - mean) / std

        x = x.reshape(b, c, s, h, w)

        return x, mean, std

    def unnorm(self, x: torch.Tensor, mean: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
        """Unnormalize the input."""
        b, c, h, w = x.shape
        input_data = x.reshape(b, self.norm_groups, -1)
        return (input_data * std + mean).reshape(b, c, h, w)

    def pad(self, x: torch.Tensor) -> Tuple[torch.Tensor, Tuple[List[int], List[int], int, int]]:
        """Pad the input with zeros to make it square.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        Tuple[torch.Tensor, Tuple[List[int], List[int], int, int]]
            Padded tensor and padding sizes.
        """
        if x.dim() == 4:
            _, _, h, w = x.shape
        elif x.dim() == 5:
            _, _, _, h, w = x.shape
        w_mult = ((w - 1) | self.padding_size) + 1
        h_mult = ((h - 1) | self.padding_size) + 1
        w_pad = [math.floor((w_mult - w) / 2), math.ceil((w_mult - w) / 2)]
        h_pad = [math.floor((h_mult - h) / 2), math.ceil((h_mult - h) / 2)]
        x = torch.nn.functional.pad(x, w_pad + h_pad)
        return x, (h_pad, w_pad, h_mult, w_mult)

    @staticmethod
    def unpad(x: torch.Tensor, h_pad: List[int], w_pad: List[int], h_mult: int, w_mult: int) -> torch.Tensor:
        """Unpad the input.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.
        h_pad : List[int]
            Height padding sizes.
        w_pad : List[int]
            Width padding sizes.
        h_mult : int
            Height multiplier.
        w_mult : int
            Width multiplier.

        Returns
        -------
        torch.Tensor
            Unpadded tensor.
        """
        return x[..., h_pad[0] : h_mult - h_pad[1], w_pad[0] : w_mult - w_pad[1]]

    @typecheck()
    def forward(self, image: torch.Tensor, **kwargs) -> torch.Tensor:  # pylint: disable=arguments-differ
        """Forward pass of :class:`BaseSegmentationNet`.

        Parameters
        ----------
        image : torch.Tensor
            Input image. Shape [batch_size, n_x, n_y] or [batch_size, n_x, n_y, 2]
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        torch.Tensor
            Predicted segmentation. Shape [batch_size, n_classes, n_x, n_y]
        """
        if self.consecutive_slices > 1:
            batch, slices = image.shape[:2]
            image = image.reshape(batch * slices, *image.shape[2:])

        if image.shape[-1] == 2:
            if self.input_channels == 1:
                image = torch.view_as_complex(image).unsqueeze(1)
                if self.magnitude_input:
                    image = torch.abs(image)
            elif self.input_channels == 2 and not self.magnitude_input:
                image = image.permute(0, 3, 1, 2)
            else:
                raise ValueError(f"The input channels must be either 1 or 2. Found: {self.input_channels}")
        elif self.magnitude_input:
            image = torch.abs(image)

        if image.dim() == 3:
            image = image.unsqueeze(1)

        mean = 1.0
        std = 1.0
        if self.normalize:
            image, mean, std = self.norm(image)
        image, pad_sizes = self.pad(image)
        segmentation = self.segmentation_module(image)
        segmentation = self.unpad(segmentation, *pad_sizes)
        if self.normalize:
            segmentation = self.unnorm(segmentation, mean, std)

        if self.normalize_segmentation_output:
            segmentation = (segmentation - segmentation.min()) / (segmentation.max() - segmentation.min())

        if self.consecutive_slices > 1:
            segmentation = segmentation.reshape(batch, slices, *segmentation.shape[1:])

        return torch.abs(segmentation)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/utils/model_utils.py

import copy
import importlib
import os
import sys
from dataclasses import dataclass, is_dataclass
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any, List, Optional, Set, Tuple, Union

import wrapt
from omegaconf import DictConfig, ListConfig, OmegaConf
from omegaconf.errors import OmegaConfBaseException
from packaging.version import Version
from pytorch_lightning import LightningModule

import atommic
from atommic.constants import ATOMMIC_ENV_CACHE_DIR
from atommic.core.classes.common import PretrainedModelInfo
from atommic.core.classes.modelPT import ModelPT
from atommic.core.conf.modelPT import atommicConfig
from atommic.utils import logging
from atommic.utils.app_state import AppState

_HAS_HYDRA = True

_VAL_TEST_FASTPATH_KEY = "ds_item"

__all__ = [
    "ArtifactPathType",
    "ArtifactItem",
    "resolve_dataset_name_from_cfg",
    "parse_dataset_as_name",
    "unique_names_check",
    "resolve_validation_dataloaders",
    "wrap_training_step",
    "convert_model_config_to_dict_config",
    "_convert_config",
    "maybe_update_config_version",
    "import_class_by_path",
    "resolve_subclass_pretrained_model_info",
    "check_lib_version",
    "resolve_cache_dir",
    "inject_model_parallel_rank",
    "uninject_model_parallel_rank",
]


class ArtifactPathType(Enum):
    """ArtifactPathType refers to the type of the path that the artifact is located at.
    LOCAL_PATH: A user local filepath that exists on the file system.
    TAR_PATH: A (generally flattened) filepath that exists inside an archive (that may have its own full path).
    """

    LOCAL_PATH = 0
    TAR_PATH = 1


@dataclass(init=False)
class ArtifactItem:
    """ArtifactItem is a dataclass that holds the information of an artifact."""

    path: str
    path_type: ArtifactPathType
    hashed_path: Optional[str] = None


def resolve_dataset_name_from_cfg(cfg: "DictConfig") -> Union[Union[str, int, Enum, float, bool, None], Any]:
    """Parses items of the provided sub-config to find the first potential key that resolves to an existing file or
    directory.

    # Fast-path Resolution
    In order to handle cases where we need to resolve items that are not paths, a fastpath key can be provided as
    defined in the global `_VAL_TEST_FASTPATH_KEY`.

    This key can be used in two ways :
    ## _VAL_TEST_FASTPATH_KEY points to another key in the config
    If this _VAL_TEST_FASTPATH_KEY points to another key in this config itself, then we assume we want to loop through
    the values of that key. This allows for any key in the config to become a fastpath key.

    Example
    -------
    validation_ds:

    .. code-block::

        splits: "val"
        ...
        <_VAL_TEST_FASTPATH_KEY>: "splits"  <-- this points to the key name "splits"

    Then we can write the following when overriding in hydra:
    ```python
    python train_file.py ... model.validation_ds.splits=[val1, val2, dev1, dev2] ...
    ```
    ## _VAL_TEST_FASTPATH_KEY itself acts as the resolved key
    If this _VAL_TEST_FASTPATH_KEY does not point to another key in the config, then it is assumed that the items of
    this key itself are used for resolution.

    Example
    -------
    validation_ds:

    .. code-block::

        <_VAL_TEST_FASTPATH_KEY>: "val"  <-- this points to the key name "splits"

    Then we can write the following when overriding in hydra:
    ```python
    python train_file.py ... model.validation_ds.<_VAL_TEST_FASTPATH_KEY>=[val1, val2, dev1, dev2] ...
    ```
    # IMPORTANT NOTE:
    It <can> potentially mismatch if there exist more than 2 valid paths, and the first path does *not* resolve the
    path of the data file (but does resolve to some other valid path). To avoid this side effect, place the data path
    as the first item on the config file.

    Parameters
    ----------
    cfg : DictConfig
        Sub-config of the config file.

    Returns
    -------
    Union[str, int, Enum, float, bool, None]
        A str representing the `key` of the config which hosts the filepath(s), or None in case path could not be
        resolved.
    """
    if _VAL_TEST_FASTPATH_KEY in cfg and cfg[_VAL_TEST_FASTPATH_KEY] is not None:
        fastpath_key = cfg[_VAL_TEST_FASTPATH_KEY]

        if isinstance(fastpath_key, str) and fastpath_key in cfg:
            return cfg[fastpath_key]
        return _VAL_TEST_FASTPATH_KEY

    for key, value in cfg.items():
        if type(value) in [list, tuple, ListConfig]:
            # Count the number of valid paths in the list
            values_are_paths = 0
            for val_i in value:
                val_i = str(val_i)

                if os.path.exists(val_i) or os.path.isdir(val_i):
                    values_are_paths += 1
                else:
                    # reset counter and break inner loop
                    break

            if values_are_paths == len(value):
                return key

        elif os.path.exists(str(value)) or os.path.isdir(str(value)):
            return key

    return None


def parse_dataset_as_name(name: str) -> str:
    """Constructs a valid prefix-name from a provided file path.

    Parameters
    ----------
    name : str
        Path to some valid data/manifest file or a python object that will be used as a name for the data loader (via
        str() cast).

    Returns
    -------
    str
        A valid prefix-name for the data loader.
    """
    name = Path(name).stem if os.path.exists(name) or os.path.isdir(name) else name
    # cleanup name
    name = name.replace("-", "_")

    if "manifest" in name:
        name = name.replace("manifest", "")

    if "dataset" in name:
        name = name.replace("dataset", "")

    # Test if the manifest/dataset name was simply `manifest.yaml` or `dataset.yaml`: Invalid names.
    if name == "":
        raise ValueError(
            "Provided dataset / manifest filename was `manifest.json` or `dataset.json`.\n"
            "Such a name is invalid, since multiple datasets/manifests can share the same name,\n"
            "thereby overriding their results during logging. Please pick a more descriptive filename \n"
            "for the provided dataset / manifest file."
        )

    if name[-1] != "_":
        name = f"{name}_"

    return name


def unique_names_check(name_list: Optional[List[str]]):
    """Performs a uniqueness check on the name list resolved, so that it can warn users about non-unique keys.

    Parameters
    ----------
    name_list : Optional[List[str]]
        List of strings resolved for data loaders.
    """
    if name_list is None:
        return

    # Name uniqueness checks
    names = set()
    for name in name_list:
        if name in names:
            logging.warning(
                "Name resolution has found more than one data loader having the same name !\n"
                "In such cases, logs will nor be properly generated. "
                "Please rename the item to have unique names.\n"
                f"Resolved name : {name}"
            )
        else:
            # we need just hash key check, value is just a placeholder
            names.add(name)


def resolve_validation_dataloaders(model: ModelPT):
    """Helper method that operates on the ModelPT class to automatically support multiple dataloaders for the
    validation set. It does so by first resolving the path to one/more data files via
    `resolve_dataset_name_from_cfg()`. If this resolution fails, it assumes the data loader is prepared to manually
    support / not support multiple data loaders and simply calls the appropriate setup method.
    If resolution succeeds:
    - Checks if provided path is to a single file or a list of files.
    If a single file is provided, simply tags that file as such and loads it via the setup method.
    If multiple files are provided:
    - Inject a new manifest path at index "i" into the resolved key.
    - Calls the appropriate setup method to set the data loader.
    - Collects the initialized data loader in a list and preserves it.
    - Once all data loaders are processed, assigns the list of loaded loaders to the ModelPT.
    - Finally, assigns a list of unique names resolved from the file paths to the ModelPT.

    Parameters
    ----------
    model : ModelPT
        ModelPT subclass, which requires >=1 Validation Dataloaders to be setup.
    """
    if not _HAS_HYDRA:
        logging.error("This function requires Hydra/OmegaConf and it was not installed.")
        sys.exit(1)
    cfg = copy.deepcopy(model._cfg)  # pylint: disable=protected-access
    dataloaders: List[Any] = []

    # process val_loss_idx
    if "val_dl_idx" in cfg.validation_ds:
        cfg = OmegaConf.to_container(cfg)
        val_dl_idx = cfg["validation_ds"].pop("val_dl_idx")
        cfg = OmegaConf.create(cfg)
    else:
        val_dl_idx = 0

    # Set val_loss_idx
    model._val_dl_idx = val_dl_idx  # pylint: disable=protected-access

    ds_key = resolve_dataset_name_from_cfg(cfg.validation_ds)

    if ds_key is None:
        logging.debug(
            f"Could not resolve file path from provided config - {cfg.validation_ds}. "
            "Disabling support for multi-dataloaders."
        )

        model.setup_validation_data(cfg.validation_ds)
        return

    ds_values = cfg.validation_ds[ds_key]

    if isinstance(ds_values, (list, tuple, ListConfig)):
        for ds_value in ds_values:
            cfg.validation_ds[ds_key] = ds_value
            model.setup_validation_data(cfg.validation_ds)
            dataloaders.append(model.validation_dl)

        model.validation_dl = dataloaders
        model.validation_names = [parse_dataset_as_name(ds) for ds in ds_values]  # type: ignore

        unique_names_check(name_list=model.validation_names)
        return
    model.setup_validation_data(cfg.validation_ds)
    model.validation_names = [parse_dataset_as_name(ds_values)]

    unique_names_check(name_list=model.validation_names)


def resolve_test_dataloaders(model: "ModelPT"):
    """Helper method that operates on the ModelPT class to automatically support multiple dataloaders for the test set.
    It does so by first resolving the path to one/more data files via `resolve_dataset_name_from_cfg()`. If this
    resolution fails, it assumes the data loader is prepared to manually support / not support multiple data loaders
    and simply calls the appropriate setup method.

    If resolution succeeds:
        Checks if provided path is to a single file or a list of files.
        If a single file is provided, simply tags that file as such and loads it via the setup method.
        If multiple files are provided:
            Inject a new manifest path at index "i" into the resolved key.
            Calls the appropriate setup method to set the data loader.
            Collects the initialized data loader in a list and preserves it.
            Once all data loaders are processed, assigns the list of loaded loaders to the ModelPT.
            Finally, assigns a list of unique names resolved from the file paths to the ModelPT.

    Parameters
    ----------
    model : ModelPT
        ModelPT subclass, which requires >=1 Test Dataloaders to be setup.
    """
    if not _HAS_HYDRA:
        logging.error("This function requires Hydra/OmegaConf and it was not installed.")
        sys.exit(1)
    cfg = copy.deepcopy(model._cfg)  # pylint: disable=protected-access
    dataloaders: List[Any] = []

    # process test_loss_idx
    if "test_dl_idx" in cfg.test_ds:
        cfg = OmegaConf.to_container(cfg)
        test_dl_idx = cfg["test_ds"].pop("test_dl_idx")
        cfg = OmegaConf.create(cfg)
    else:
        test_dl_idx = 0

    # Set val_loss_idx
    model._test_dl_idx = test_dl_idx  # pylint: disable=protected-access

    ds_key = resolve_dataset_name_from_cfg(cfg.test_ds)

    if ds_key is None:
        logging.debug(
            f"Could not resolve file path from provided config - {cfg.test_ds}. "
            "Disabling support for multi-dataloaders."
        )

        model.setup_test_data(cfg.test_ds)
        return

    ds_values = cfg.test_ds[ds_key]

    if isinstance(ds_values, (list, tuple, ListConfig)):
        for ds_value in ds_values:
            cfg.test_ds[ds_key] = ds_value
            model.setup_test_data(cfg.test_ds)
            dataloaders.append(model.test_dl)

        model.test_dl = dataloaders
        model.test_names = [parse_dataset_as_name(ds) for ds in ds_values]  # type: ignore

        unique_names_check(name_list=model.test_names)
        return
    model.setup_test_data(cfg.test_ds)
    model.test_names = [parse_dataset_as_name(ds_values)]

    unique_names_check(name_list=model.test_names)


@wrapt.decorator
def wrap_training_step(wrapped, instance: LightningModule, args, kwargs):
    """Wraps the training step of the LightningModule.

    Parameters
    ----------
    wrapped : function
        The wrapped function.
    instance : LightningModule
        The LightningModule instance.
    args : tuple
        The arguments passed to the wrapped function.
    kwargs : dict
        The keyword arguments passed to the wrapped function.

    Returns
    -------
    function
        The return value of the wrapped function.
    """
    output_dict = wrapped(*args, **kwargs)

    if isinstance(output_dict, dict) and output_dict is not None and "log" in output_dict:
        log_dict = output_dict.pop("log")
        instance.log_dict(log_dict, on_step=True)

    return output_dict


def convert_model_config_to_dict_config(cfg: Union[DictConfig, atommicConfig]) -> DictConfig:
    """Converts its input into a standard DictConfig.

    Possible input values are:
        - DictConfig
        - A dataclass which is a subclass of atommicConfig

    Parameters
    ----------
    cfg : Union[DictConfig, atommicConfig]
        A dict-like object.

    Returns
    -------
    DictConfig
        The equivalent DictConfig.
    """
    if not _HAS_HYDRA:
        logging.error("This function requires Hydra/OmegaConf and it was not installed.")
        sys.exit(1)
    if not isinstance(cfg, (OmegaConf, DictConfig)) and is_dataclass(cfg):
        cfg = OmegaConf.structured(cfg)

    if not isinstance(cfg, DictConfig):
        raise ValueError(f"cfg constructor argument must be of type DictConfig/dict but got {type(cfg)} instead.")

    config = OmegaConf.to_container(cfg, resolve=True)
    config = OmegaConf.create(config)
    return config


def _convert_config(cfg: "OmegaConf"):
    """Recursive function converting the configuration from old hydra format to the new one."""
    if not _HAS_HYDRA:
        logging.error("This function requires Hydra/OmegaConf and it was not installed.")
        sys.exit(1)

    # Get rid of params.
    if "params" in cfg:
        params = cfg.pop("params")
        for param_key, param_val in params.items():
            cfg[param_key] = param_val

    # Recursion.
    try:
        for _, sub_cfg in cfg.items():
            if isinstance(sub_cfg, DictConfig):
                _convert_config(sub_cfg)
    except OmegaConfBaseException as e:
        logging.warning(f"Skipped conversion for config/subconfig:\n{cfg}\n Reason: {e}.")


def maybe_update_config_version(cfg: "DictConfig"):
    """Recursively convert Hydra 0.x configs to Hydra 1.x configs.
    Changes include:
    -   `cls` -> `_target_`.
    -   `params` -> drop params and shift all arguments to parent.
    -   `target` -> `_target_` cannot be performed due to ModelPT injecting `target` inside class.

    Parameters
    ----------
    cfg : DictConfig
        Any Hydra compatible DictConfig

    Returns
    -------
    DictConfig
        An updated DictConfig that conforms to Hydra 1.x format.
    """
    if not _HAS_HYDRA:
        logging.error("This function requires Hydra/OmegaConf and it was not installed.")
        sys.exit(1)
    if cfg is not None and not isinstance(cfg, DictConfig):
        try:
            temp_cfg = OmegaConf.create(cfg)
            cfg = temp_cfg
        except OmegaConfBaseException:
            # Cannot be cast to DictConfig, skip updating.
            return cfg

    # Make a copy of model config.
    cfg = copy.deepcopy(cfg)
    OmegaConf.set_struct(cfg, False)

    # Convert config.
    _convert_config(cfg)

    # Update model config.
    OmegaConf.set_struct(cfg, True)

    return cfg


@lru_cache(maxsize=1024)
def import_class_by_path(path: str):
    """Recursive import of class by path string."""
    paths = path.split(".")
    path = ".".join(paths[:-1])
    class_name = paths[-1]
    mod = __import__(path, fromlist=[class_name])
    mod = getattr(mod, class_name)
    return mod


def resolve_subclass_pretrained_model_info(base_class) -> Union[List[PretrainedModelInfo], Set[Any]]:
    """Recursively traverses the inheritance graph of subclasses to extract all pretrained model info.
    First constructs a set of unique pretrained model info by performing DFS over the inheritance graph. All model info
     belonging to the same class is added together.

    Parameters
    ----------
    base_class : class
        The root class, whose subclass graph will be traversed.

    Returns
    -------
    list
        A list of unique pretrained model infos belonging to all the inherited subclasses of this baseclass.
    """
    list_of_models = set()

    def recursive_subclass_walk(cls):
        """
        Recursively traverses the inheritance graph of subclasses to extract all pretrained model info.

        Parameters
        ----------
        cls : class
            The class to be traversed.

        Returns
        -------
        list
            A list of unique pretrained model infos belonging to all the inherited subclasses of this baseclass.
        """
        for subclass in cls.__subclasses__():
            # step into its immediate subclass
            recursive_subclass_walk(subclass)

            subclass_models = subclass.list_available_models()

            if subclass_models is not None and len(subclass_models) > 0:
                # Inject subclass info into pretrained model info, if not already overridden by subclass.
                for model_info in subclass_models:
                    # If subclass manually injects class_, dont override.
                    if model_info.class_ is None:
                        model_info.class_ = subclass

                for model_info in subclass_models:
                    list_of_models.add(model_info)

    recursive_subclass_walk(base_class)
    list_of_models = list(sorted(list_of_models))  # type: ignore
    return list_of_models


def check_lib_version(lib_name: str, checked_version: str, operator) -> Tuple[Optional[bool], str]:
    """Checks if a library is installed, and if it is, checks the operator(lib.__version__, checked_version) as a
    result. This bool result along with a string backend of result is returned. If the library is not installed at all,
     then returns None instead, along with a string explaining that the library is not installed

    Parameters
    ----------
    lib_name : str
        Lower case str name of the library that must be imported.
    checked_version : str
        Semver string that is compared against lib.__version__.
    operator : bool
        Binary callable function func(a, b) -> bool; that compares lib.__version__ against version in some manner. Must
         return a boolean.

    Returns
    -------
    tuple
        A tuple of results:
            -   Bool or None. Bool if the library could be imported, and the result of
                operator(lib.__version__, checked_version) or False if __version__ is not implemented in lib.
                None is passed if the library is not installed at all.
            -   A string analysis of the check.
    """
    try:
        if "." in lib_name:
            mod = import_class_by_path(lib_name)
        else:
            mod = importlib.import_module(lib_name)

        if hasattr(mod, "__version__"):
            lib_ver = Version(mod.__version__)
            match_ver = Version(checked_version)

            if operator(lib_ver, match_ver):
                msg = f"Lib {lib_name} version is satisfied !"
                return True, msg
            msg = (
                f"Lib {lib_name} version ({lib_ver}) is not {operator.__name__} than required version "
                f"{checked_version}.\n"
                "Please upgrade the lib using either pip or conda to the latest version."
            )
            return False, msg
        msg = (
            f"Lib {lib_name} does not implement __version__ in its init file. "
            "Could not check version compatibility."
        )
        return False, msg
    except (AttributeError, ImportError, ModuleNotFoundError):
        pass

    msg = f"Lib {lib_name} has not been installed. Please use pip or conda to install this package."
    return None, msg


def resolve_cache_dir() -> Path:
    """Utility method to resolve a cache directory for atommic that can be overridden by an environment variable.

    Example
    -------
        atommic_CACHE_DIR="~/atommic_cache_dir/" python atommic_example_script.py

    Returns
    -------
    A Path object, resolved to the absolute path of the cache directory. If no override is provided, uses an inbuilt
    default which adapts to atommic versions strings.
    """
    override_dir = os.environ.get(ATOMMIC_ENV_CACHE_DIR, "")
    return (
        Path.joinpath(Path.home(), f".cache/torch/atommic/atommic_{atommic.__version__}")
        if override_dir == ""
        else Path(override_dir).resolve()
    )


def uninject_model_parallel_rank(filepath):
    """Uninjects tensor/pipeline model parallel ranks from the filepath."""
    filepath = str(filepath)
    if "mp_rank" in filepath or "tp_rank" in filepath:
        dirname = os.path.dirname(os.path.dirname(filepath))
        basename = os.path.basename(filepath)
        filepath = os.path.join(dirname, basename)
    return filepath


def inject_model_parallel_rank(filepath):
    """Injects tensor/pipeline model parallel ranks into the filepath. Does nothing if not using model parallelism."""
    filepath = uninject_model_parallel_rank(filepath)
    app_state = AppState()
    if app_state.model_parallel_size is not None and app_state.model_parallel_size > 1:
        # filepath needs to be updated to include mp_rank
        dirname = os.path.dirname(filepath)
        basename = os.path.basename(filepath)
        if app_state.pipeline_model_parallel_size is None or app_state.pipeline_model_parallel_size == 1:
            filepath = f"{dirname}/mp_rank_{app_state.tensor_model_parallel_rank:02d}/{basename}"
        else:
            filepath = (
                f"{dirname}/tp_rank_{app_state.tensor_model_parallel_rank:02d}_pp_rank_"
                f"{app_state.pipeline_model_parallel_rank:03d}/{basename} "
            )
        return filepath
    return filepath


def ckpt_to_dir(filepath: Union[str, Path]) -> Path:
    """PTL considers checkpoints as .ckpt files. This method removes the extension and returns a path to be used as a
    directory for distributed checkpoints.
    """
    filepath = Path(filepath)
    # adding this assert because we will later remove directories based on the return value of this method
    assert filepath.suffix == ".ckpt", f'filepath: {filepath} must have .ckpt extension'
    # create a new path whose name is the original filepath without the .ckpt extension
    checkpoint_dir = filepath.with_name(filepath.stem)
    return checkpoint_dir
